/*
 * Pluginsystem
 * Copyright (C) Thasso Griebel
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package org.java.plugin.extension;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.java.plugin.JpfException;
import org.java.plugin.extension.annotations.scanner.ParameterContainer;
import org.java.plugin.registry.*;
import org.java.plugin.registry.ExtensionPoint.ParameterDefinition;
import org.java.plugin.registry.IntegrityCheckReport.ReportItem;
import org.java.plugin.registry.xml.IntegrityChecker;

import java.util.*;


public class AnnotatedExtension implements Extension{
	private Log log = LogFactory.getLog(getClass());
	private String id;
	private PluginDescriptor pluginDescriptor;
	private String extendedPluginID;
	private String extendedPointID;
	private List<Parameter> parameters;
	private Boolean isValid;
     
	public AnnotatedExtension(PluginDescriptor pluginDescriptor, String id, String plugin,
			String point, List<ParameterContainer> parameters) {
		this.id = id;
		this.extendedPluginID = plugin;
		this.extendedPointID = point;
		this.pluginDescriptor = pluginDescriptor;
		this.parameters = new ArrayList<Parameter>();
		for (ParameterContainer p : parameters) {
			this.parameters.add(new AnnotatedParameter(this, pluginDescriptor, null, p.getId(), p.getValue(), p.getType()));
		}
		
	}
	public String getExtendedPluginId() {
		return extendedPluginID;
	}

	public String getExtendedPointId() {
		return extendedPointID;
	}

	public Parameter getParameter(String id) {
        Parameter result = null;
        for (Parameter param : parameters) {            
            if (param.getId().equals(id)) {
                if (result == null) {
                    result = param;
                } else {
                    throw new IllegalArgumentException(
                        "more than one parameter with ID " + id //$NON-NLS-1$
                        + " defined in extension " + getUniqueId()); //$NON-NLS-1$
                }
            }
        }
        return result;

	}

	public Collection<Parameter> getParameters() {
		return parameters;
	}

	public Collection<Parameter> getParameters(String id) {
        List<Parameter> result = new LinkedList<Parameter>();
        for (Parameter parameter : parameters) {
            if (parameter.getId().equals(id))
                result.add(parameter);
        }
        return Collections.unmodifiableList(result);
	}

    public String getUniqueId() {
        return getDeclaringPluginDescriptor().getRegistry().makeUniqueId(
                getDeclaringPluginDescriptor().getId(), getId());
    }

	public String getId() {
		return id;
	}

	public PluginDescriptor getDeclaringPluginDescriptor() {
		return pluginDescriptor;
	}

	public PluginFragment getDeclaringPluginFragment() {
		return null;
	}

	public String getDocsPath() {
		return null;
	}

	public Documentation<Extension> getDocumentation() {
		return null;
	}
	
    ExtensionPoint getExtensionPoint(final String uniqueId) {
        PluginRegistry registry = getDeclaringPluginDescriptor().getRegistry();
        return getExtensionPoint(registry.extractPluginId(uniqueId),
                registry.extractId(uniqueId));
    }

    ExtensionPoint getExtensionPoint(final String pluginId,
            final String pointId) {
        PluginRegistry registry = getDeclaringPluginDescriptor().getRegistry();
        if (!registry.isPluginDescriptorAvailable(pluginId)) {
            return null;
        }
        for (ExtensionPoint point : registry.getPluginDescriptor(pluginId)
                .getExtensionPoints()) {
            if (point.getId().equals(pointId)) {
                return point;
            }
        }
        return null;
    }

    public boolean isValid() {
        if (isValid == null) {
            validate();
        }
        return isValid.booleanValue();
    }

    public String get(String name) {
        Parameter p = getParameter(name);
        if(p != null){
            return p.rawValue();
        }else{
           // check for a default value
           ExtensionPoint ep = getExtensionPoint(getExtendedPluginId(), getExtendedPointId());
           if(ep.getParameterDefinition(name) != null){
               return ep.getParameterDefinition(name).getDefaultValue();
           }
        }
        return null;        
    }

    public <T> T get(String name, Class<? extends T> type) {
        String value = get(name);
        if(value == null) return null;
        InstanceFactory f = InstanceFactory.getInstance();
        return (T)f.convert(value, String.class, type, f.getPluginClassLoader(this));        
    }

    // made public to fix #14
    public Collection<ReportItem> validate() {
        ExtensionPoint point =
            getExtensionPoint(getExtendedPluginId(), getExtendedPointId());
        if (point == null) {
            isValid = Boolean.FALSE;
            return Collections.singletonList((ReportItem)
                    new IntegrityChecker.ReportItemImpl(
                        IntegrityCheckReport.Severity.ERROR, this,
                        IntegrityCheckReport.Error.INVALID_EXTENSION,
                        "extPointNotAvailable", new Object[] { //$NON-NLS-1$
                        getDeclaringPluginDescriptor().getRegistry()
                            .makeUniqueId(getExtendedPluginId(),
                                    getExtendedPointId()), getUniqueId()}));
        }
        Collection<ReportItem> result =
            validateParameters(point.getParameterDefinitions(), parameters);
        isValid = result.isEmpty() ? Boolean.TRUE : Boolean.FALSE;
        return result;
    }
    
    private Collection<ReportItem> validateParameters(final Collection<ParameterDefinition> allDefinitions,
            final Collection<Parameter> allParams) {
        List<ReportItem> result = new LinkedList<ReportItem>();
        Map<String, Collection<Parameter>> groups = new HashMap<String, Collection<Parameter>>();
        for (Parameter param : allParams) {
            ParameterDefinition def = param.getDefinition();
            if (def == null) {
                result.add(new IntegrityChecker.ReportItemImpl(
                        IntegrityCheckReport.Severity.ERROR, this,
                        IntegrityCheckReport.Error.INVALID_EXTENSION,
                        "cantDetectParameterDef", new Object[] { //$NON-NLS-1$
                        param.getId(), getUniqueId()}));
                continue;
            }
            if (groups.containsKey(param.getId())) {
                groups.get(param.getId()).add(param);
            } else {
                Collection<Parameter> paramGroup = new LinkedList<Parameter>();
                paramGroup.add(param);
                groups.put(param.getId(), paramGroup);
            }
        }
        if (!result.isEmpty()) {
            return result;
        }

        List<Parameter> empty_paramGroup = Collections.emptyList();
        for (ParameterDefinition def : allDefinitions) {
            Collection<Parameter> paramGroup = groups.get(def.getId());
            result.addAll(validateParameters(def,
                    (paramGroup != null) ? paramGroup : empty_paramGroup));
        }
        return result;
    }
    
    private Collection<ReportItem> validateParameters(final ParameterDefinition def,
            final Collection<Parameter> params) {
        if (log.isDebugEnabled()) {
            log.debug("validating parameters for definition " + def); //$NON-NLS-1$
        }
        switch (def.getMultiplicity()) {
        case ONE:
            if (params.size() != 1) {
                return Collections.singletonList((ReportItem)
                        new IntegrityChecker.ReportItemImpl(
                                IntegrityCheckReport.Severity.ERROR, this,
                                IntegrityCheckReport.Error.INVALID_EXTENSION,
                                "tooManyOrFewParams", new Object[] { //$NON-NLS-1$
                                        def.getId(), getUniqueId()}));
            }
            break;
        case NONE_OR_ONE:
            if (params.size() > 1) {
                return Collections.singletonList((ReportItem)
                        new IntegrityChecker.ReportItemImpl(
                                IntegrityCheckReport.Severity.ERROR, this,
                                IntegrityCheckReport.Error.INVALID_EXTENSION,
                                "tooManyParams", new Object[] { //$NON-NLS-1$
                                        def.getId(), getUniqueId()}));
            }
            break;
        case ONE_OR_MORE:
            if (params.isEmpty()) {
                return Collections.singletonList((ReportItem)
                        new IntegrityChecker.ReportItemImpl(
                                IntegrityCheckReport.Severity.ERROR, this,
                                IntegrityCheckReport.Error.INVALID_EXTENSION,
                                "tooFewParams", new Object[] { //$NON-NLS-1$
                                        def.getId(), getUniqueId()}));
            }
            break;
        case ANY:
            // no-op
            break;
        }
        if (params.isEmpty()) {
            return Collections.emptyList();
        }
        List<ReportItem> result = new LinkedList<ReportItem>();
        int count = 1;
        AnnotatedParameter param;
        for (Parameter parameter : params) {
            param = (AnnotatedParameter) parameter;
            if (!param.isValid()) {
                result.add(new IntegrityChecker.ReportItemImpl(
                        IntegrityCheckReport.Severity.ERROR, this,
                        IntegrityCheckReport.Error.INVALID_EXTENSION,
                        "invalidParameterValue", new Object[] { //$NON-NLS-1$
                        def.getId(), Integer.valueOf(count), getUniqueId()}));
            }
            if ((ParameterType.ANY != def.getType()) && result.isEmpty()) {
                result.addAll(validateParameters(
                        param.getDefinition().getSubDefinitions(),
                        param.getSubParameters()));
            }
            count++;	// FIXME in 0.11 not in 0.12
        }
        return result;
    }

    @Override
    public String toString() {
        return "{AnnotatedPluginExtension: uid=" + getUniqueId() + "}"; //$NON-NLS-1$ //$NON-NLS-2$
    }

    public Object createInstance() throws JpfException {
        return InstanceFactory.getInstance().createFromExtension(this);
    }
}
