/*
 * Pluginsystem
 * Copyright (C) Thasso Griebel
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package org.java.plugin.extension;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.java.plugin.PathResolver;
import org.java.plugin.registry.*;
import org.java.plugin.registry.Extension.Parameter;
import org.java.plugin.registry.ExtensionPoint.ParameterDefinition;
import org.java.plugin.registry.xml.ExtensionPointImpl.ParameterDefinitionImpl;
import org.java.plugin.registry.xml.ParameterValueParser;

import java.net.URL;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;

/// annotated parameter implementation
public class AnnotatedParameter implements Parameter {
	private Log log = LogFactory.getLog(getClass());
    private ParameterValueParser valueParser;
    private ParameterDefinition definition = null;
    private boolean definitionDetected = false;
	private String id;
	private String value;
	private Class<?> type;
	
	private Extension extension;
	private PluginDescriptor plugin;
	private PluginFragment fragment;

    public AnnotatedParameter(
			Extension extension,
			PluginDescriptor plugin,
			PluginFragment fragment,
			String id, String value, Class<?> type) {
    	this.extension = extension;
    	this.plugin = plugin;
    	this.fragment = fragment;
    	this.id = id;
    	this.value = value;
    	this.type = type;
	}

	/**
     * @see org.java.plugin.registry.Extension.Parameter#getDeclaringExtension()
     */
    public Extension getDeclaringExtension() {
        return extension;
    }

    /**
     * @see org.java.plugin.registry.PluginElement#getDeclaringPluginDescriptor()
     */
    public PluginDescriptor getDeclaringPluginDescriptor() {
        return plugin;
    }
    
    /**
     * @see org.java.plugin.registry.PluginElement#getDeclaringPluginFragment()
     */
    public PluginFragment getDeclaringPluginFragment() {
        return fragment;
    }

    /**
     * @see org.java.plugin.registry.Extension.Parameter#getDefinition()
     */
    public ParameterDefinition getDefinition() {
        if (definitionDetected) {
            return definition;
        }
        definitionDetected = true;
        if (log.isDebugEnabled()) {
            log.debug("detecting definition for parameter " + this); //$NON-NLS-1$
        }
        Collection<ParameterDefinition> definitions;
        definitions = getExtensionPoint(
                    getDeclaringExtension().getExtendedPluginId(),
                    getDeclaringExtension().getExtendedPointId()).
                        getParameterDefinitions();
        for (ParameterDefinition def : definitions) {
            if (def.getId().equals(getId())) {
                definition = def;
                break;
            }
        }
        if (log.isDebugEnabled()) {
            log.debug("definition detected - " + definition); //$NON-NLS-1$
        }
        return definition;
    }
    
    ExtensionPoint getExtensionPoint(final String pluginId,
            final String pointId) {
        PluginRegistry registry = getDeclaringPluginDescriptor().getRegistry();
        if (!registry.isPluginDescriptorAvailable(pluginId)) {
            return null;
        }
        for (ExtensionPoint point : registry.getPluginDescriptor(pluginId)
                .getExtensionPoints()) {
            if (point.getId().equals(pointId)) {
                return point;
            }
        }
        return null;
    }

    /**
     * @see org.java.plugin.registry.Extension.Parameter#getSuperParameter()
     */
    public Parameter getSuperParameter() {
        return null;
    }
    
    /**
     * @see org.java.plugin.registry.Extension.Parameter#getSubParameters()
     */
    @SuppressWarnings("unchecked")
	public Collection<Parameter> getSubParameters() {
        return Collections.EMPTY_LIST;
    }
    
    /**
     * @see org.java.plugin.registry.Extension.Parameter#getSubParameter(
     *      java.lang.String)
     */
    public Parameter getSubParameter(final String id) {
        return null;
    }

    /**
     * @see org.java.plugin.registry.Extension.Parameter#getSubParameters(
     *      java.lang.String)
     */
    @SuppressWarnings("unchecked")
	public Collection<Parameter> getSubParameters(final String id) {
        return Collections.EMPTY_LIST;
    }

    /**
     * @see org.java.plugin.registry.Extension.Parameter#rawValue()
     */
    public String rawValue() {
        return (value != null) ? value : "";
    }
    
    boolean isValid() {
        if (valueParser != null) {
            return valueParser.isParsingSucceeds();
        }
        if (log.isDebugEnabled()) {
            log.debug("validating parameter " + this); //$NON-NLS-1$
        }
        valueParser = new ParameterValueParser(
                getDeclaringPluginDescriptor().getRegistry(),
                getDefinition(), value);
        if (!valueParser.isParsingSucceeds()) {
            log.warn("parsing value for parameter " + this //$NON-NLS-1$
                    + " failed, message is: " //$NON-NLS-1$
                    + valueParser.getParsingMessage());
        }
        return valueParser.isParsingSucceeds();
    }
    
    /**
     * @see org.java.plugin.registry.Extension.Parameter#valueAsBoolean()
     */
    public Boolean valueAsBoolean() {
        if (!isValid()) {
            throw new UnsupportedOperationException(
                    "parameter value is invalid"); //$NON-NLS-1$
        }
        if (ParameterType.BOOLEAN != definition.getType()) {
            throw new UnsupportedOperationException(
                    "parameter type is not " //$NON-NLS-1$
                    + ParameterType.BOOLEAN);
        }
        if (valueParser.getValue() == null) {
            return (Boolean) ((ParameterDefinitionImpl) getDefinition()).getValueParser().getValue();
        }
        return (Boolean) valueParser.getValue();
    }
    
    /**
     * @see org.java.plugin.registry.Extension.Parameter#valueAsDate()
     */
    public Date valueAsDate() {
        if (!isValid()) {
            throw new UnsupportedOperationException(
                    "parameter value is invalid"); //$NON-NLS-1$
        }
        if ((ParameterType.DATE != definition.getType())
                && (ParameterType.DATE_TIME != definition.getType())
                && (ParameterType.TIME != definition.getType())) {
            throw new UnsupportedOperationException("parameter type is not " //$NON-NLS-1$
                    + ParameterType.DATE + " nor " //$NON-NLS-1$
                    + ParameterType.DATE_TIME + " nor" //$NON-NLS-1$
                    + ParameterType.TIME);
        }
        if (valueParser.getValue() == null) {
            return (Date) ((ParameterDefinitionImpl) getDefinition())
            .getValueParser().getValue();
        }
        return (Date) valueParser.getValue();
    }
    
    /**
     * @see org.java.plugin.registry.Extension.Parameter#valueAsNumber()
     */
    public Number valueAsNumber() {
        if (!isValid()) {
            throw new UnsupportedOperationException(
                    "parameter value is invalid"); //$NON-NLS-1$
        }
        if (ParameterType.NUMBER != definition.getType()) {
            throw new UnsupportedOperationException(
                    "parameter type is not " //$NON-NLS-1$
                    + ParameterType.NUMBER);
        }
        if (valueParser.getValue() == null) {
            return (Number) ((ParameterDefinitionImpl) getDefinition())
            .getValueParser().getValue();
        }
        return (Number) valueParser.getValue();
    }
    
    /**
     * @see org.java.plugin.registry.Extension.Parameter#valueAsString()
     */
    public String valueAsString() {
        if (!isValid()) {
            throw new UnsupportedOperationException(
                    "parameter value is invalid"); //$NON-NLS-1$
        }
        if ((ParameterType.STRING != definition.getType())
                && (ParameterType.FIXED != definition.getType())) {
            throw new UnsupportedOperationException(
                    "parameter type is not " //$NON-NLS-1$
                    + ParameterType.STRING);
        }
        if (valueParser.getValue() == null) {
            return (String) ((ParameterDefinitionImpl) getDefinition())
            .getValueParser().getValue();
        }
        return (String) valueParser.getValue();
    }

    /**
     * @see org.java.plugin.registry.Extension.Parameter#valueAsExtension()
     */
    public Extension valueAsExtension() {
        if (!isValid()) {
            throw new UnsupportedOperationException(
                    "parameter value is invalid"); //$NON-NLS-1$
        }
        if (ParameterType.EXTENSION_ID != definition.getType()) {
            throw new UnsupportedOperationException(
                    "parameter type is not " //$NON-NLS-1$
                    + ParameterType.EXTENSION_ID);
        }
        if (valueParser.getValue() == null) {
            return (Extension) ((ParameterDefinitionImpl) getDefinition())
            .getValueParser().getValue();
        }
        return (Extension) valueParser.getValue();
    }
    
    /**
     * @see org.java.plugin.registry.Extension.Parameter#valueAsExtensionPoint()
     */
    public ExtensionPoint valueAsExtensionPoint() {
        if (!isValid()) {
            throw new UnsupportedOperationException(
                    "parameter value is invalid"); //$NON-NLS-1$
        }
        if (ParameterType.EXTENSION_POINT_ID != definition.getType()) {
            throw new UnsupportedOperationException(
                    "parameter type is not " //$NON-NLS-1$
                    + ParameterType.EXTENSION_POINT_ID);
        }
        if (valueParser.getValue() == null) {
            return (ExtensionPoint) (
                    (ParameterDefinitionImpl) getDefinition())
                    .getValueParser().getValue();
        }
        return (ExtensionPoint) valueParser.getValue();
    }
    
    /**
     * @see org.java.plugin.registry.Extension.Parameter#valueAsPluginDescriptor()
     */
    public PluginDescriptor valueAsPluginDescriptor() {
        if (!isValid()) {
            throw new UnsupportedOperationException(
                    "parameter value is invalid"); //$NON-NLS-1$
        }
        if (ParameterType.PLUGIN_ID != definition.getType()) {
            throw new UnsupportedOperationException("parameter type is not " //$NON-NLS-1$
                    + ParameterType.PLUGIN_ID);
        }
        if (valueParser.getValue() == null) {
            return (PluginDescriptor) (
                    (ParameterDefinitionImpl) getDefinition())
                    .getValueParser().getValue();
        }
        return (PluginDescriptor) valueParser.getValue();
    }
    
    /**
     * @see org.java.plugin.registry.Extension.Parameter#valueAsUrl()
     */
    public URL valueAsUrl() {
        return valueAsUrl(null);
    }

    /**
     * @see org.java.plugin.registry.Extension.Parameter#valueAsUrl(
     *      org.java.plugin.PathResolver)
     */
    public URL valueAsUrl(final PathResolver pathResolver) {
        if (!isValid()) {
            throw new UnsupportedOperationException(
                    "parameter value is invalid"); //$NON-NLS-1$
        }
        if (ParameterType.RESOURCE != definition.getType()) {
            throw new UnsupportedOperationException(
                    "parameter type is not " //$NON-NLS-1$
                    + ParameterType.RESOURCE);
        }
        if ((valueParser.getValue() == null) && (rawValue() == null)) {
            return valueAsUrl(pathResolver,
                    getDefinition().getDeclaringExtensionPoint(),
                    (URL) ((ParameterDefinitionImpl) getDefinition())
                    .getValueParser().getValue(),
                    getDefinition().getDefaultValue());
        }
        return valueAsUrl(pathResolver, getDeclaringPluginDescriptor(),
                (URL) valueParser.getValue(), rawValue());
    }
    
    private URL valueAsUrl(final PathResolver pathResolver,
            final Identity idt, final URL absoluteUrl,
            final String relativeUrl) {
        if ((pathResolver == null) || (absoluteUrl != null)) {
            return absoluteUrl;
        }
        if (relativeUrl == null) {
            return null;
        }
        return pathResolver.resolvePath(idt, relativeUrl);
    }
    
    /**
     * @see java.lang.Object#toString()
     */
    @Override
    public String toString() {
        return "{PluginExtension.Parameter: extUid=" //$NON-NLS-1$
            + getDeclaringExtension().getUniqueId() + "; id=" + getId() //$NON-NLS-1$
            + "}"; //$NON-NLS-1$
    }

	public String getId() {
		return id;
	}

	public String getDocsPath() {
		return null;
	}

	public Documentation<Parameter> getDocumentation() {
		return null;
	}
}

