/*
 * Pluginsystem
 * Copyright (C) Thasso Griebel
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package org.java.plugin.extension;

import org.java.plugin.PluginManager;

import java.net.URL;
import java.net.MalformedURLException;

/**
 * Simple internal plugin location. Both manifest and context location point to the given
 * URL. The URL you pass to the constructor must point to a valid XML file that defines a plugin.
 *
 * @author Thasso Griebel (thasso.griebel@gmail.com)
 */
public class InternalPluginLocation implements PluginManager.PluginLocation{
    /**
     * Manifest location
     */
    private URL pluginXML;
    /**
     * the context URL
     */
    private URL contextURL;

    /**
     * Create a new Internal location. The url must point to a valid Plugin XML file.
     *
     * @param pluginXML url to the plugin file
     */
    public InternalPluginLocation(URL pluginXML){
        super();
        if(pluginXML == null){
            throw new NullPointerException("Null PLugin location is not allowed!");
        }
        this.pluginXML = pluginXML;
        
        // build the context url
        try {
            String urlString = convertToContext(pluginXML.toExternalForm());
            this.contextURL = new URL(urlString);
        } catch (MalformedURLException e) {
            throw new IllegalArgumentException("Unable to create context location for " + pluginXML);
        }
    }

    public URL getManifestLocation() {
        return pluginXML;
    }

    /**
     * The context is unused here and  points to the same URL as the Manifest.
     *
     * @return url to context (in this case unused and equal to the manifest URL)
     */
    public URL getContextLocation() {
        return contextURL;
    }

    /**
     * Helper method that removes the plugin file from the path. This also removes META-INF.
     *
     * @param urlString
     * @return path to the plugins context folder 
     */
    protected String convertToContext(String urlString){
        if(urlString.contains("META-INF")){
            return urlString.substring(0, urlString.lastIndexOf("META-INF"));
        }else{
            return urlString.substring(0, urlString.lastIndexOf("/")+1);    
        }
    }
}
