/*
 * Pluginsystem
 * Copyright (C) Thasso Griebel
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package org.java.plugin.extension.annotations.scanner;

import org.java.plugin.extension.annotations.scanner.archiveiterator.Filter;
import org.java.plugin.extension.annotations.scanner.archiveiterator.IteratorFactory;
import org.java.plugin.extension.annotations.scanner.archiveiterator.StreamIterator;
import org.java.plugin.registry.Extension;
import org.java.plugin.registry.PluginDescriptor;
import org.objectweb.asm.*;

import java.io.*;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import java.util.Stack;


/**
 *
 *
 * The class allows you to scan an arbitrary set of "archives" for .class files.  These class files
 * are parsed to see what annotations they use.  Two indexes are created.
 *
 * One is a map of annotations and what classes
 * use those annotations.   This could be used, for example, by an EJB deployer to find all the EJBs contained
 * in the archive
 *
 * Another is a mpa of classes and what annotations those classes use.
 *
 * Code is boroght from
 * <a href="http://bill.burkecentral.com/2008/01/14/scanning-java-annotations-at-runtime/">
 * http://bill.burkecentral.com/2008/01/14/scanning-java-annotations-at-runtime/
 * </a>
 * @author <a href="mailto:bill@burkecentral.com">Bill Burke</a>
 * @author Thasso Griebel (Thasso.Griebel@gmail.com)
 * @version $Revision: 1 $
 */
public class AnnotationDB implements Serializable
{
    private PluginDescriptor plugin;
    private List<Extension> extensionList;
    private boolean loaded = false;


    public AnnotationDB(PluginDescriptor plugin, List<Extension> extensions) {
        this.plugin = plugin;
        this.extensionList = extensions;
        loaded = true;
    }

	public AnnotationDB(PluginDescriptor plugin) {
    	this.plugin = plugin;
    	this.extensionList = new ArrayList<Extension>();
	}
    
	public List<Extension> getExtensions(){
		return extensionList;
	}
    
	/**
    * Scan a url that represents an "archive"  this is a classpath directory or jar file
    *
    * @param urls variable list of URLs to scan as archives
    * @throws IOException
    */
   public void scanArchives(URL... urls) throws IOException
   {
      for (URL url : urls)
      {
         Filter filter = new Filter()
         {
            public boolean accepts(String filename)
            {
               return filename.endsWith(".class");
            }
         };

         StreamIterator it = IteratorFactory.create(url, filter);
         InputStream stream;
         while ((stream = it.next()) != null) scanClass(stream);
      }

   }

   /**
    * Parse a .class file for annotations
    *
    * @param bits input stream pointing to .class file bits
    * @throws IOException
    */
   public void scanClass(InputStream bits) throws IOException
   {	   
      DataInputStream dstream = new DataInputStream(new BufferedInputStream(bits));
      ClassReader reader = new ClassReader(dstream);
      try
      {
          reader.accept(new ClassCrawler(), ClassReader.SKIP_DEBUG);
      }
      finally
      {
         dstream.close();
         bits.close();
      }
   }
   
   /**
    * Basic {@link ClassAdapter} to crawl a plugin class. It checks for Annotations
    * and delegates calls to the appropriate visitor implementation.
    * 
    * @author thasso
    *
    */
   class ClassCrawler extends ClassAdapter{	   
	   private Stack<String> classNames = new Stack<String>();
	   
	   public ClassCrawler() {
		   super(new EmptyVisitor());
	   }
	   	
		@Override
		public void visit(int version, int access, String name, String signature,
			String superName, String[] interfaces) {			
			classNames.push(toClassName(name));
			super.visit(version, access, name, signature, superName, interfaces);
		}

		@Override
		public AnnotationVisitor visitAnnotation(String desc, boolean visible) {
			String annotationClassName = Type.getType(desc).getClassName();
			if(annotationClassName.equals(org.java.plugin.extension.annotations.Extension.class.getName())){
				return new ExtensionVisitor(plugin, classNames.peek(), extensionList);
			}
			return null;
		}
				
		@Override
		public void visitEnd() {
			super.visitEnd();
			classNames.pop();
		}

		private String toClassName(String s){
			return s.replace("/", ".");
		}
   }

    public boolean isLoaded() {
        return loaded;
    }

    /**
    * Helper at the end of the visitor chain.
    * 
    * @author thasso
    *
    */
   class EmptyVisitor implements ClassVisitor{
	public void visit(int version, int access, String name, String signature,
			String superName, String[] interfaces) {}
	public AnnotationVisitor visitAnnotation(String desc, boolean visible) {
		return null;
	}
	public void visitAttribute(Attribute attr) {}
	public void visitEnd() {}
	public FieldVisitor visitField(int access, String name, String desc,
			String signature, Object value) {
		return null;}
	public void visitInnerClass(String name, String outerName,
			String innerName, int access) {}
	public MethodVisitor visitMethod(int access, String name, String desc,
			String signature, String[] exceptions) {
		return null;}
	public void visitOuterClass(String owner, String name, String desc) {}
	public void visitSource(String source, String debug) {}	   
   }
}
