/*
 * Pluginsystem
 * Copyright (C) Thasso Griebel
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package org.java.plugin.extension.util;

import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Enumeration;
import java.util.List;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;

import org.java.plugin.registry.PluginDescriptor;
import org.java.plugin.util.IoUtil;

public class PluginIOUtils {
	/**
	 * Takes the {@link PluginDescriptor} and uses its
	 * {@link PluginDescriptor#getLocation()} method to find the plugins
	 * location. The location URL is reduced to its path part and the reduced
	 * URL is returned. The default location points to a plugin.xml in the
	 * plugins root and the path reduction should point to that root.
	 * 
	 * @param desc
	 * @return url to the plugins root
	 */
	public static URL getPluginPath(PluginDescriptor desc) {
		URL location = desc.getLocation();
		if (location == null)
			return null;
		String url = location.toExternalForm();
		String path = url.substring(0, url.length() - "plugin.xml".length());
		try {
			URL pathURL = new URL(path);
			return pathURL;
		} catch (MalformedURLException ignore) {
		}
		return null;
	}

	public static URL getPluginPathTo(PluginDescriptor desc, String path) {
		URL location = desc.getLocation();
		if (location == null)
			return null;
		String url = location.toExternalForm();
        String p = url;
        if(p.endsWith("/plugin.xml")){
		    p = url.substring(0, url.length() - "plugin.xml".length());
        }else{
            int lastSlash = p.lastIndexOf("/");
            if(lastSlash > 0){
                p = url.substring(0, lastSlash);
            }else{
                p = url;
            }
        }

        // cut out META-INF
        if(p.endsWith("META-INF")){
            p = p.substring(0, p.length() - "META-INF".length());
        }

		if(!p.endsWith("/") && !path.startsWith("/")) p = p+"/";
		if(p.endsWith("/") && path.startsWith("/")) path = path.substring(1);
		p = p+path;
        
		try {
			URL pathURL = new URL(p);
			return pathURL;
		} catch (MalformedURLException ignore) {
		}
		return null;

	}

	/**
	 * Returns true if the given {@link PluginDescriptor} contains the resource,
	 * that is, the {@link IoUtil#isResourceExists(URL)} returns true for the
	 * URL created from the plugins location and the given resource.
	 * 
	 * @param pluginDescriptor
	 * @param resource
	 * @return true if resource exists
	 */
	public static boolean isResourceAvailable(
			PluginDescriptor pluginDescriptor, String resource) {
		if (resource.startsWith("/")){
			resource = resource.substring(1);
		}
		URL pluginPath = getPluginPath(pluginDescriptor);
		String root = null;
		if (pluginPath != null){
			root = pluginPath.toExternalForm();			
		}
		try {			
			URL location = new URL(root + (root.endsWith("/") ? "":"/") + resource);			
			return location != null && IoUtil.isResourceExists(location);
		} catch (MalformedURLException ignore) {			
		}
		return false;
	}

	public static Collection<String> getAllElementsIn(PluginDescriptor plugin,
			String path) {
		if (path.startsWith("/"))
			path = path.substring(1);
		URL pluginPath = getPluginPath(plugin);
		if ("jar".equalsIgnoreCase(pluginPath.getProtocol())) {
			return getAllElementsInJar(pluginPath, path);
		} else {
			return getAllElementsInDir(pluginPath, path);
		}		
	}

	protected static Collection<String> getAllElementsInDir(URL url, String path) {
		File pluginDir = IoUtil.url2file(url);
		if(pluginDir == null) return Collections.EMPTY_LIST;
		File f = new File(pluginDir, path);
		if(f == null) return Collections.EMPTY_LIST;
		if(!f.canRead() || !f.isDirectory()) return Collections.EMPTY_LIST;		
		ArrayList<File> all = new ArrayList<File>();
		File[]  files = f.listFiles();
		for (File file : files) {
			if(file.isFile()) all.add(file);
			else addFilesFromDir(file, all);
		}
		ArrayList<String> fnames = new ArrayList<String>();
		int p = pluginDir.getAbsolutePath().length();
		for (File file : all) {			
			fnames.add(file.getAbsolutePath().substring(p+1));
		}
		return fnames;
	}
	
	private static void addFilesFromDir(File file, ArrayList<File> all) {
		if(!file.isDirectory()) return;
		File[] files = file.listFiles();
		for (File f : files) {
			if(f.isFile()) all.add(f);
			else addFilesFromDir(f, all);
		}
	}

	protected static Collection<String> getAllElementsInJar(URL url, String path) {
		String urlStr = url.toExternalForm();
		int p = urlStr.indexOf("!/"); //$NON-NLS-1$
		if (p == -1) {// this is invalid JAR file URL
			return Collections.EMPTY_LIST;
		}
		URL fileUrl = null;
		try {
			fileUrl = new URL(urlStr.substring(4, p));
		} catch (MalformedURLException e2) {
			// TODO Auto-generated catch block
			e2.printStackTrace();
		}
		if(fileUrl == null){
			// TODO: throw exception here
			return Collections.EMPTY_LIST;			
		}
		File file = IoUtil.url2file(fileUrl);
		if (file == null) {// this is non-local JAR file URL
			// TODO: throw exception here
			return Collections.EMPTY_LIST;
		}
		if (!file.canRead()) {
			// TODO: throw exception here
			return Collections.EMPTY_LIST;
		}
		JarFile jarFile = null;
		try {
			jarFile = new JarFile(file);
			List<String> entries = new ArrayList<String>();
			Enumeration<JarEntry> entryEnum = jarFile.entries();			
			while (entryEnum.hasMoreElements()) {
				JarEntry e = entryEnum.nextElement();
				if (e.getName().startsWith(path) && !e.isDirectory()) {
					entries.add(e.getName());
				}
			}
			jarFile.close();
			return entries;
		} catch (IOException e1) {
			// TODO: handle error here
		} finally {
			if (jarFile != null) {
				try {
					jarFile.close();
				} catch (Exception ignore) {
				}
			}
		}
		return null;
	}

}
